//==============================================================================
// PRINTF - Some printf/sprintf type of functions.  Also includes writing out
//		to basic atom pro IO ports.
//
// Version 0.1 - Work in Progress
//
// Major Contributers: Kurt Eckhardt 
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//==============================================================================
#include "BapLib.h"
#include <stdarg.h>
#include "hsio.h"

#ifdef __cplusplus
extern "C"
{
#endif

byte	g_bSOPin;
word	g_wSOMode;
PORTITEM g_piTSOPin;							// The port item for the specified pin

static void printchar(char **str, int c)
{
	extern int putchar(int c);
	if (str) {
		**str = c;
		++(*str);
	} else { 
		char b;
		b = (char)c;
		
		if (g_bSOPin == (byte)_S_HSERIAL_) 
			hserout(1, (u8*)&b,1);
#ifndef BASICATOMPRO28
		else if (g_bSOPin == (byte)_S_HSERIAL2_) 
			hserout(2, (u8*)&b,1);
#endif		
		else
			TASerout(g_bSOPin, g_wSOMode, (const char*)&b, 1);
	}
}

#define PAD_RIGHT 1
#define PAD_ZERO 2

static int prints(char **out, const char *string, int width, int pad)
{
	register int pc = 0, padchar = ' ';

	if (width > 0) {
		register int len = 0;
		register const char *ptr;
		for (ptr = string; *ptr; ++ptr) ++len;
		if (len >= width) width = 0;
		else width -= len;
		if (pad & PAD_ZERO) padchar = '0';
	}
	if (!(pad & PAD_RIGHT)) {
		for ( ; width > 0; --width) {
			printchar (out, padchar);
			++pc;
		}
	}
	for ( ; *string ; ++string) {
		printchar (out, *string);
		++pc;
	}
	for ( ; width > 0; --width) {
		printchar (out, padchar);
		++pc;
	}

	return pc;
}

/* the following should be enough for 32 bit int */
#define PRINT_BUF_LEN 12

static int printi(char **out, long int i, int b, int sg, int width, int pad, int letbase, int fLong)
{
	char print_buf[PRINT_BUF_LEN];
	register char *s;
	register int t, neg = 0, pc = 0;
	register unsigned long int u;			// = i;

	
	// In this version we are trying to handle both long and short integers.  On Short Integers the upper word may be garbage...
	if (fLong) {
		u = i;
		if (sg && b == 10 && i < 0) {
			neg = 1;
			u = -i;
		}
	} else {
		u = i & 0xffff;		// only process the lower 16 bits
		if (sg && b == 10 && (u & 0x8000)) {
			short si = (short)u;		// 
			neg = 1;
			u = -si;					// Hopefully this took care of the conversion back
		} 
	}

	// If the number is zero output this quickly...	
	if (u == 0) {
		print_buf[0] = '0';
		print_buf[1] = '\0';
		return prints (out, print_buf, width, pad);
	}

	s = print_buf + PRINT_BUF_LEN-1;
	*s = '\0';

	while (u) {
		t = u % b;
		if( t >= 10 )
			t += letbase - '0' - 10;
		*--s = t + '0';
		u /= b;
	}

	if (neg) {
		if( width && (pad & PAD_ZERO) ) {
			printchar (out, '-');
			++pc;
			--width;
		} else {
			*--s = '-';
		}
	}

	return pc + prints (out, s, width, pad);
}

static int print(char **out, const char* format, va_list ap)
{
	register int width, pad, fLong;
	register int pc = 0;
	char scr[2];

	for (; *format != 0; ++format) {
		if (*format == '%') {
			++format;
			width = pad = 0;
			fLong = 0;
			if (*format == '\0') break;
			if (*format == '%') goto out;
			if (*format == '-') {
				++format;
				pad = PAD_RIGHT;
			}
			while (*format == '0') {
				++format;
				pad |= PAD_ZERO;
			}
			for ( ; *format >= '0' && *format <= '9'; ++format) {
				width *= 10;
				width += *format - '0';
			}
			if (*format == 'l') {
				++format;
				fLong = 1;
			}
			if( *format == 's' ) {
				register char *s = va_arg(ap, char *);
				pc += prints (out, s?s:"(null)", width, pad);
				continue;
			}
			if( *format == 'd' ) {
				pc += printi (out, va_arg(ap, long int), 10, 1, width, pad, 'a', fLong);
				continue;
			}
			if( *format == 'x' ) {
				pc += printi (out, va_arg(ap,long int), 16, 0, width, pad, 'a', fLong);
				continue;
			}
			if( *format == 'X' ) {
				pc += printi (out, va_arg(ap,long int), 16, 0, width, pad, 'A', fLong);
				continue;
			}
			if( *format == 'u' ) {
				pc += printi (out, va_arg(ap,long int), 10, 0, width, pad, 'a', fLong);
				continue;
			}
			if( *format == 'c' ) {
				/* char are converted to int then pushed on the stack */
				scr[0] = (char)va_arg(ap,int);
				scr[1] = '\0';
				pc += prints (out, scr, width, pad);
				continue;
			}
		} else {
		out:
			printchar (out, *format);
			++pc;
		}
	}
	if (out) **out = '\0';
	va_end(ap);
	return pc;
}

/* assuming sizeof(void *) == sizeof(int) */

int printf(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);

	return print(0, format, ap);
}

int lsprintf(char *out, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	return print(&out, format, ap);
}

int lvsprintf(char *out, const char *format, va_list ap)
{
	return print(&out, format, ap);
}


int IOPrintf(byte bPin, word wMode, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);

	g_bSOPin = bPin;
	g_wSOMode = wMode;
	return print(0, format, ap);
}

int IOVPrintf(byte bPin, word wMode, const char *format, va_list ap)
{
	g_bSOPin = bPin;
	g_wSOMode = wMode;
	return print(0, format, ap);
}

int DBGPrintf(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
#ifdef BASICATOMPRO28
	g_bSOPin = _S_OUT_;
	g_wSOMode = I9600;
#else
	g_bSOPin = (u8)-1;	// Use hserout for our debug on other platforms...

#endif	
	return print(0, format, ap);
}

void WaitForDBGTransmitComplete(void)
{
#ifndef BASICATOMPRO28
	// Only need to do something if we are a hardware serial port
	while(!hserstat(1, HSERSTAT_OUTPUT_EMPTY))
		;
#endif
}



//==================================================================================================
//==================================================================================================
// Timer Based Serial Input and Output Functions.
//==================================================================================================
//==================================================================================================

//==============================================================================
// _TAS_INITTIMER - Internal function to do the initialize the timer based on
//			input baud mode. - Note the baud mode is assumed to be the 
// CPU_SPEED/(baud * 8)
//==============================================================================
byte _TAS_INITTIMER(word wMode)
{
	// BUGBUG: should we use the BASIC form of baud rate here?
	// ok need to initialize TimerV
	bool	fInvert;
	word	wBitTime;
	byte	bClkDiv;
	
	fInvert = ((wMode & (1<<14)) != 0);
	
	wBitTime = wMode & 0x1fff;				// Extract off the bit time
	bClkDiv = 3;					// start off assuming clock / 8	
	while (wBitTime >= 0x100) {		// we need to fit into a byte...
		bClkDiv++;					// increment to the next clock divisor
		wBitTime >>= 1;				// shift it down one bit
	}

	// Now lets set the Timers registers.
	TV.TCORA = wBitTime -1;			// save away the time to start...
	TV.TCRV1.BYTE = bClkDiv & 0x1;		// The low order bit goes into TCRV1
	TV.TCRV0.BYTE = (bClkDiv >> 1) | 0x8;	// setup the other bits of the timer

	return fInvert? 0xff : 0;
}


#define WAIT_TIMER_COMPARE_MATCH 		{while (!TV.TCSRV.BIT.CMFA) ;	TV.TCSRV.BIT.CMFA = 0;}

//==============================================================================
// [TASerout(pin, baudmode, buffer, bufferlen]
//
// Second attempt will be to use TimerV which is only 8 bits, this version will
// use the Compare Match A interrupt do do the acutual output bit processing.
// The main loop will still wait in for the work to be done.  Could later change
// this to queue the values and do it all in the background, but that could get
// complicated as maybe I would start doing multiple different IO ports at the same
// time...  But maybe later.
//
//==============================================================================

void TASerout(byte bPin, word wMode, const char *pb, byte cb)
{
	byte bInvertMask;						// do we need to invert everything
	byte b;									// next byte to output	
	byte	cBits;							// bit counter
	
	bInvertMask = _TAS_INITTIMER(wMode);	// Init Timer V for this operation.

	// Use the port table to get the IOPins PDR register and mask use for speed...
	g_piTSOPin = (*g_pPortTable)[bPin];		// get the port table item for the specified pin...

	// Make sure the IO line is correct to start off with
	if (bInvertMask) {
		PI_HIGH(g_piTSOPin);
	} else {
		PI_LOW(g_piTSOPin);
	}	
	output(bPin);		// make sure the port is in the right state...
	
	// Now setup data pointer and counter.
	WAIT_TIMER_COMPARE_MATCH

	// Wait at least one timer unit before we start
	WAIT_TIMER_COMPARE_MATCH			// wait for timer match
	
	//;------------------------------------------------
	//; Main loop to output our data.  The IO line should
	//; be initialized and should be in the stop state...
	//;
	//; Current registers are:
	//; r2	- Word address for PDRx for the output pin
	//; r3l	- bit number of IO PIN
	//; r3h	- 0 if Normal communication, FF if inverted.
	//; e3	- Count of bytes left to output	
	//; er4 	- Pointing to the current byte to output

	while (cb > 0) {
		b = (*pb++) ^ bInvertMask;		// get the next byte to output and take care of Xor or not...
		cBits = 8;						// count of bits to output
	
		// Wait for start bit.
		WAIT_TIMER_COMPARE_MATCH		// wait for the next clock.
		
		if (bInvertMask) {
			PI_LOW(g_piTSOPin);
		} else {
			PI_HIGH(g_piTSOPin);	
		}

		while (cBits > 0) {
			WAIT_TIMER_COMPARE_MATCH		// wait for the next timer
			if (b & 0x1) {
				PI_LOW(g_piTSOPin);
			} else {
				PI_HIGH(g_piTSOPin);
			}
			b >>= 1;						// shift it down one bit
			cBits--;
		}

		// Now lets do the stop bit
		WAIT_TIMER_COMPARE_MATCH			// wait for the next timer
		if (bInvertMask) {
			PI_HIGH(g_piTSOPin);	
		} else {
			PI_LOW(g_piTSOPin);
		}			

		cb--;
	}

}

//==============================================================================
// [TASerin(pin, fPin, baudmode, buffer, bufferlen, timeout, wEOL]
//	This is a timer based input function.  This one will also try to handle
//	some flow control and handle up to two extra bytes of data being sent
//	after we toggle the flow control pin as for example the XBEE may send
//	one or two characters at that point...
//   Also it can optionally exit when an optional end character <> 0 is read
//
//
//==============================================================================
byte	g_bTasiOBuf[4];			// Used for flow control overrun...
byte 	g_bTASI_cFO;

byte TASerin(byte bPin, byte bfPin, word wMode, char *pb, byte cb, unsigned long lTimeout, word wEOL)
{
#ifdef DBG_TASERIN
	PORTITEM piDBG;							// used for debug
#endif	
	byte bInvertMask;						// do we need to invert everything
	byte b;									// next byte to output	
	byte	cBits;							// bit counter
	byte	bTcntvAtSB;						// 
	byte	cbIn;							// count of bytes asked for	
	unsigned long 	lTO;					// timeout
	bInvertMask = _TAS_INITTIMER(wMode);	// Init Timer V for this operation.
	cbIn = cb; 	

	input (bPin);							// make sure BPin is set to input
	// Use the port table to get the IOPins PDR register and mask use for speed...
	g_piTSOPin = (*g_pPortTable)[bPin];		// get the port table item for the specified pin...

	// debug
#ifdef DBG_TASERIN
	piDBG = (*g_pPortTable)[0];				// setup to debug our outputs
	output(0);
#endif
		
	// BUGBUG:: first pass not processing flow pin...
	bTcntvAtSB	= (TV.TCORA * 3) / 4;

	// convert Timout to approx how many cycles through our loop we should do...
	if (lTimeout != 0xffffffff)
		lTimeout /= 10;						// First guess of how many uS per loop
	
	while (cb > 0) {
		b = 0;
		// When we get the first edge of the start bit, we want to adjust the Timers Count
		// such that we should get overflow near the middle of the start pulse.  Preferably
		// a slight be earlier then center
		// When we have the start we want to setup our timer to be half way through
		// a bit cycle to read in the center of it.

		// First we need to wait for the start bit to start
		// BUGBUG:: Put in time out here...
		lTO = lTimeout;
		if (bInvertMask) {
			// we are inverted - wait until we have a start bit
			while (PI_ISHIGH(g_piTSOPin)) {
				if (lTO != 0xffffffff) {
					lTO--;
					if (lTO == 0)
						goto _TASI_TO;
				}
			}
		} else {
			// Not inverted...
			while (!PI_ISHIGH(g_piTSOPin)) {
				if (lTO != 0xffffffff){
					lTO--;
					if (lTO == 0)
						goto _TASI_TO;
				}
			}			
		}
		
		// We have the stop bit
		TV.TCNTV = bTcntvAtSB;		// set the clock to are precalculated best spot...
		TV.TCSRV.BIT.CMFA = 0;		// Make sure the overflow is off
#ifdef DBG_TASERIN
		PI_TOGGLE(piDBG);
#endif
		cBits = 8;
		
		// First timer should now be in the start bit
		WAIT_TIMER_COMPARE_MATCH				//

		while (cBits > 0) {
			WAIT_TIMER_COMPARE_MATCH				// wait for the next timer
			b = (b >> 1) | (PI_ISHIGH(g_piTSOPin) ? 0 : 0x80);				// process that bit

#ifdef DBG_TASERIN
			PI_TOGGLE(piDBG);
#endif
			cBits--;
		}
		
		b ^=  bInvertMask;							// If necessary do invert our bits...
		*pb++ = b;									// save away the retrieved byte
		cb--;
		WAIT_TIMER_COMPARE_MATCH					// wait for the next stop bit
#ifdef DBG_TASERIN
		PI_TOGGLE(piDBG);
#endif
		if (b == wEOL)
			break;
	}
_TASI_TO:
	return cbIn - cb;								// return how many bytes were retrieved.				
}

//==============================================================================
// [TASClearFOBuf - clear any input bytes that we cached.
//==============================================================================
void TASClearFOBuf(void)
{
	g_bTASI_cFO = 0;		// simply zero the count
}

#ifdef __cplusplus
}  
#endif

